#!/usr/bin/env python

"""
Common classes.

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.data import Attr, Instance, UnresolvedName
from micropython.errors import *
import sys

try:
    set
except NameError:
    from sets import Set as set

# Visitors and activities related to node annotations.

class ASTVisitor:

    "A base class for visitors."

    def default(self, node, *args):
        for n in node.getChildNodes():
            self.dispatch(n)

    def dispatch(self, node, *args):

        "Dispatch using 'node', annotating any raised exceptions."

        # Dispatch via a generic visit method.

        try:
            return node.visit(self, *args)

        # Annotate the exception in case of failure.

        except NodeProcessingError, exc:
            if exc.astnode is None:
                exc.astnode = node

                # NOTE: Should perhaps specialise the subclasses appropriately.

                if hasattr(self, "unit"):
                    exc.unit_name = self.unit.full_name()
                else:
                    exc.unit_name = self.full_name()
            raise

    def possible_accessor_types(self, node, defining_users=1):

        """
        Given annotations made during the inspection process, return all possible
        type names and indications of static usage for a 'node' involved in
        attribute access.

        If 'defining_users' is set to a false value, attempt to get the type
        names specifically applicable to the node, rather than retrieving more
        general definition-based type observations.
        """

        target_names = set()

        if hasattr(node, "_attr") and not isinstance(node._attr, (Instance, UnresolvedName)):
            attr = node._attr
            if isinstance(attr, Attr):
                target_names.add((attr.parent.full_name(), attr.is_static_attribute()))
            else:
                target_names.add((attr.full_name(), attr.is_static_attribute()))

        elif hasattr(node, "_attrusers"):

            # Visit each attribute user.

            for user in node._attrusers:

                # Since users such as branches may not provide type information,
                # attempt to find defining users.

                if defining_users:
                    for def_user in user._attrdefs or [user]:
                        for target_name, is_static in def_user._attrtypes.get(node._username, []):
                            target_names.add((target_name, is_static))
                else:
                    for target_name, is_static in user._attrspecifictypes.get(node._username, []):
                        target_names.add((target_name, is_static))

        return target_names

def used_by_unit(node):

    """
    Return whether the definition made by a 'node' is actually employed by the
    program unit within which it is found.
    """

    return hasattr(node, "unit") and node.unit.parent.has_key(node.unit.name)

# Useful data.

operator_functions = {

    # Binary operations.

    "Add" : "add",
    "Bitand" : "and_",
    "Bitor" : "or_",
    "Bitxor" : "xor",
    "Div" : "div",
    "FloorDiv" : "floordiv",
    "LeftShift" : "lshift",
    "Mod" : "mod",
    "Mul" : "mul",
    "Power" : "pow",
    "RightShift" : "rshift",
    "Sub" : "sub",

    # Unary operations.

    "Invert" : "invert",
    "UnaryAdd" : "pos",
    "UnarySub" : "neg",

    # Augmented assignment.

    "+=" : "iadd",
    "-=" : "isub",
    "*=" : "imul",
    "/=" : "idiv",
    "//=" : "ifloordiv",
    "%=" : "imod",
    "**=" : "ipow",
    "<<=" : "ilshift",
    ">>=" : "irshift",
    "&=" : "iand",
    "^=" : "ixor",
    "|=" : "ior",

    # Comparisons.

    "==" : "eq",
    "!=" : "ne",
    "<" : "lt",
    "<=" : "le",
    ">=" : "ge",
    ">" : "gt",

    # Access and slicing.

    "AssSlice" : "setslice",
    "Slice" : "getslice",
    "AssSubscript" : "setitem",
    "Subscript" : "getitem",
    }

# vim: tabstop=4 expandtab shiftwidth=4
